use std::{collections::VecDeque, sync::Arc};

use crate::{
    commands::device_profile::version::{deserialize_version, DESERIALIZE_VERSION_RAW_NAME},
    enums::ProtocolVersion,
};

use super::{
    command_base::{CommandDeserializer, DeserializedCommand},
    device_profile::{
        mix_effect_block_config::{
            MixEffectBlockConfigDeserializer, DESERIALIZE_MIX_EFFECT_BLOCK_CONFIG_NAME,
        },
        product_identifier::{
            ProductIdentifierDeserializer, DESERIALIZE_PRODUCT_IDENTIFIER_RAW_NAME,
        },
        topology::{TopologyDeserializer, DESERIALIZE_TOPOLOGY_RAW_NAME},
    },
    init_complete::{InitCompleteDeserializer, DESERIALIZE_INIT_COMPLETE_RAW_NAME},
    mix_effects::program_input::{ProgramInputDeserializer, DESERIALIZE_PROGRAM_INPUT_RAW_NAME},
    tally_by_source::{TallyBySourceDeserializer, DESERIALIZE_TALLY_BY_SOURCE_RAW_NAME},
    time::{TimeDeserializer, DESERIALIZE_TIME_RAW_NAME},
};

pub fn deserialize_commands(
    payload: &[u8],
    version: &mut ProtocolVersion,
) -> VecDeque<Arc<dyn DeserializedCommand>> {
    let mut parsed_commands: VecDeque<Arc<dyn DeserializedCommand>> = VecDeque::new();
    let mut head = 0;

    while payload.len() > head + 8 {
        let length = u16::from_be_bytes([payload[head], payload[head + 1]]) as usize;
        let Ok(name) = String::from_utf8(payload[(head + 4)..(head + 8)].to_vec()) else {
            break;
        };

        if length < 8 {
            break;
        }

        log::debug!("Received command {} with length {}", name, length);

        let command_buffer = &payload[head + 8..head + length];

        if name == DESERIALIZE_VERSION_RAW_NAME {
            let version_command = deserialize_version(command_buffer);
            *version = version_command.version.clone();
            log::info!("Switched to protocol version {}", version);
            parsed_commands.push_back(Arc::new(version_command));
        } else if let Some(deserializer) = command_deserializer_from_string(name.as_str()) {
            let deserialized_command = deserializer.deserialize(command_buffer, version);
            log::debug!("Received {:?}", deserialized_command);
            parsed_commands.push_back(deserialized_command);
        } else {
            log::warn!("Received command {name} for which there is no deserializer.");
            // TODO: Remove!
            todo!("Write deserializer for {name}.");
        }

        head += length;
    }

    parsed_commands
}

fn command_deserializer_from_string(command_str: &str) -> Option<Box<dyn CommandDeserializer>> {
    match command_str {
        DESERIALIZE_INIT_COMPLETE_RAW_NAME => Some(Box::<InitCompleteDeserializer>::default()),
        DESERIALIZE_PROGRAM_INPUT_RAW_NAME => Some(Box::<ProgramInputDeserializer>::default()),
        DESERIALIZE_TALLY_BY_SOURCE_RAW_NAME => Some(Box::<TallyBySourceDeserializer>::default()),
        DESERIALIZE_TIME_RAW_NAME => Some(Box::<TimeDeserializer>::default()),
        DESERIALIZE_TOPOLOGY_RAW_NAME => Some(Box::<TopologyDeserializer>::default()),
        DESERIALIZE_MIX_EFFECT_BLOCK_CONFIG_NAME => {
            Some(Box::<MixEffectBlockConfigDeserializer>::default())
        }
        DESERIALIZE_PRODUCT_IDENTIFIER_RAW_NAME => {
            Some(Box::<ProductIdentifierDeserializer>::default())
        }
        _ => None,
    }
}
